import glob
import json
import subprocess
import sys

#import util

def run_hooks(context, name):
    for path in _find_hooks(context, name):
        _run_hook(context, path)

def run_values_hooks(context, name):
    for path in _find_hooks(context, name):
        path = _run_values_hook(context, path)
        context['values'] = util.merge_yaml(context['values'], path)

def _find_hooks(context, name):
    hooks = []
    for folder in context['hook_dirs']:
        hooks += glob.glob('{}/**/{}*'.format(folder, name), recursive=True)

    return hooks

def _run_hook(context, path):
    data = json.dumps(context['values'])
    subprocess.call([path, data], stderr=sys.stderr.buffer, stdout=sys.stdout.buffer)

def _run_values_hook(context, path):
    data = json.dumps(context['values'])
    out = subprocess.check_output([path, data])
    return json.loads(out)


import click
import yaml
import os

import imka

@click.group()
@click.option('--imka-context', type=str, help='specify the imka config context')
@click.option('--imka-config', type=str, default=os.path.expanduser('~/.config/d4rk.io/imka/config.yml'), help='specify the imka config path')
@click.pass_context
def main(ctx, imka_context, imka_config):
    ctx.obj = imka.init_context()

    ctx.obj['options']['imka_config'] = imka_config
    ctx.obj['options']['imka_context'] = imka_context

@main.command()
@click.argument('frame', type=str)
@click.argument('deployment', type=str)
@click.option('--values', '-f', multiple=True, type=click.Path(exists=True), help='specify values in YAML files to customize the frame deployment')
@click.option('--render-values-depth', type=int, default=32, help='specify the max allowed value template nesteding depth')
@click.pass_context
def values(ctx, frame, deployment, values, render_values_depth):
    imka.load_values(ctx.obj, frame, deployment, values, render_values_depth)

    print(yaml.dump(ctx.obj['values']))

@main.command()
@click.argument('frame', type=str)
@click.argument('deployment', type=str)
@click.option('--values', '-f', multiple=True, type=click.Path(exists=True), help='specify values in YAML files to customize the frame deployment')
@click.option('--render-values-depth', type=int, default=32, help='specify the max allowed value template nesteding depth')
@click.pass_context
def template(ctx, frame, deployment, values, render_values_depth):
    imka.render_templates(ctx.obj, frame, deployment, values, render_values_depth)

    print('---')
    print(yaml.dump(ctx.obj['docker_stack_yml']))

@main.command()
@click.argument('frame', type=str)
@click.argument('deployment', type=str)
@click.option('--values', '-f', multiple=True, type=click.Path(exists=True), help='specify values in YAML files to customize the frame deployment')
@click.option('--render-values-depth', type=int, default=32, help='specify the max allowed value template nesteding depth')
@click.option('--prune-config-versions', is_flag=True, help='specify that old config versions should be delete')
@click.option('--prune-mount-versions', is_flag=True, help='specify that old mount versions should be delete')
@click.pass_context
def apply(ctx, frame, deployment, values, render_values_depth, prune_config_versions, prune_mount_versions):
    imka.apply(ctx.obj, frame, deployment, values, render_values_depth, prune_config_versions, prune_mount_versions)

@main.command()
@click.argument('frame', type=str)
@click.argument('deployment', type=str)
@click.option('--values', '-f', multiple=True, type=click.Path(exists=True), help='specify values in YAML files to customize the frame deployment')
@click.option('--render-values-depth', type=int, default=32, help='specify the max allowed value template nesteding depth')
@click.pass_context
def down(ctx, frame, deployment, values, render_values_depth):
    imka.down(ctx.obj, frame, deployment, values, render_values_depth)

@main.group()
def context():
    pass

@context.command(name='list')
@click.pass_context
def list_command(ctx):
    imka.context_list(ctx.obj)

@context.command(name='show')
@click.argument('context', type=str)
@click.pass_context
def context_show(ctx, context):
    imka.context_show(ctx.obj, context)

@context.command()
@click.argument('context', type=str)
@click.pass_context
def use(ctx, context):
    imka.context_use(ctx.obj, context)


import docker
import os

def open_with_context(context, path, mode='r'):
    return open(get_real_path(context, path), mode)

def get_real_path(context, path):
    return os.path.join(context['options']['frame_base_path'], path)

def merge_yaml(destination, source):
    for key, value in source.items():
        if isinstance(value, dict) and value:
            node = destination.get(key, {})
            destination[key] = merge_yaml(node, value)
        else:
            destination[key] = value

    return destination

def init_context():
    context = {}
    context['options'] = {}
    context['value_files'] = []
    context['values'] = {}
    context['configs'] = []
    context['mounts'] = []
    context['hook_dirs'] = []
    
    context['docker_client'] = docker.from_env()

    return context


import jinja2
import json
import yaml
import copy

from . import util
from .configs import Config
from .mounts import Mount


class ImkaFunctions:
    def __init__(self, context):
        self.context = context

    def config_from_file(self, path, docker_templating=False):
        config = Config.from_file(self.context, path, docker_templating)

        return json.dumps({"external": True, "name": config.version})

    def config_from_template(self, path):
        with util.open_with_context(self.context, path) as file:
            template = file.read()

        content = render_template(self.context, template)

        config = Config.from_content(self.context, path, content)

        return json.dumps({"external": True, "name": config.version})

    def mount_dir(self, path):
        mount = Mount.from_path(self.context, path)

        return mount.mountPath


def render_template(context, content):
    j2context = copy.copy(context['values'])
    j2context['imka'] = ImkaFunctions(context)


    template = jinja2.Template(content)
    rendered = template.render(j2context)

    return yaml.safe_load(rendered)

def render_compose_templates(context):
    compose = {}

    for path in context['compose_templates']:
        with util.open_with_context(context, path) as file:
            content = file.read()
        
        compose = util.merge_yaml(compose, render_template(context, content))

        print('compose_template {} rendered'.format(path))

    context['docker_stack_yml'] = compose


import yaml
import sys

from subprocess import Popen, PIPE, STDOUT

def stack_apply(context):
    name = context['values']['deployment_fullname']

    dockerStack = context['docker_stack_yml']
    dockerStackYaml = yaml.dump(dockerStack)
    dockerStackYamlEncoded = dockerStackYaml.encode()

    p = Popen(['docker', 'stack', 'deploy', '-c', '-', name], stdin=PIPE, stderr=sys.stderr.buffer, stdout=sys.stdout.buffer)
    p.communicate(input=dockerStackYamlEncoded)[0]

def stack_down(context):
    name = context['values']['deployment_fullname']

    p = Popen(['docker', 'stack', 'rm', name], stderr=sys.stderr.buffer, stdout=sys.stdout.buffer)
    p.communicate()[0]


import os
import yaml

import imka.frame
import imka.values

from . import imka_config
from . import hooks 
from . import templates
from . import configs
from . import mounts
from . import stack

def load_values(context, frame, deployment, values, render_values_depth):
    imka_config.load_system_config(context)

    context['options']['frame'] = frame

    imka.frame.load_frame(context)

    context['values']['deployment'] = deployment
    context['values']['deployment_fullname'] = '{}-{}'.format(context['values']['frame_name'], context['values']['deployment'])

    context['options']['render_values_depth'] = render_values_depth
    context['value_files'] += values

    hooks.run_values_hooks(context, 'pre-values')

    imka.values.load_values(context)

    hooks.run_values_hooks(context, 'post-values')

def render_templates(context, frame, deployment, values, render_values_depth):
    load_values(context, frame, deployment, values, render_values_depth)

    templates.render_compose_templates(context)

def apply(context, frame, deployment, values, render_values_depth, prune_config_versions, prune_mount_versions):
    context['options']['remove_old_config_versions_on_apply'] = prune_config_versions
    context['options']['remove_old_mount_versions_on_apply'] = prune_mount_versions

    render_templates(context, frame, deployment, values, render_values_depth)

    hooks.run_hooks(context, 'pre-apply')

    configs.apply_configs(context)
    mounts.apply_mounts(context)

    stack.stack_apply(context)

    configs.after_apply_configs(context)
    mounts.after_apply_mounts(context)

    hooks.run_hooks(context, 'post-apply')

def down(context, frame, deployment, values, render_values_depth):
    load_values(context, frame, deployment, values, render_values_depth)

    hooks.run_hooks(context, 'pre-down')

    stack.stack_down(context)

    configs.down_configs(context)
    mounts.down_mounts(context)

    hooks.run_hooks(context, 'post-down')

def context_list(context):
    if os.path.exists(context['options']['imka_config']):
        
        with open(context['options']['imka_config']) as file:
            imkaConfig = yaml.safe_load(file)

        for key in imkaConfig['contexts'].keys():
            print(key)

def context_show(context, name):
    if os.path.exists(context['options']['imka_config']):
        with open(context['options']['imka_config']) as file:
            imkaConfig = yaml.safe_load(file)

        if name in imkaConfig['contexts']:
            print(yaml.dump(imkaConfig['contexts'][name]))

def context_use(context, name):
    if os.path.exists(context['options']['imka_config']):
        with open(context['options']['imka_config']) as file:
            imkaConfig = yaml.safe_load(file)

        if name not in imkaConfig['contexts']:
            print("error context dose not exist")
        
        imkaConfig['context'] = name

        with open(context['options']['imka_config'], 'w') as file:
            file.write(yaml.dump(imkaConfig))


import jinja2
import yaml

from . import util

def load_values(context):
    read_from_files(context)
    render(context)

def read_from_files(context):
    values = context['values']

    for path in context['value_files']:
        with open(path) as file:
            current = yaml.safe_load(file)
        values = util.merge_yaml(values, current)

    context['values'] = values


def render(context):
    depth = context['options'].get('render_values_depth', 32)
    values = context['values']

    for i in range(depth):
        values = _render_values_step(values, values, i == depth-1)

    context['values'] = values

def _render_values_step(values, node, last):
    new = {}

    for key, value in node.items():
        if isinstance(value, dict) and value:
            new[key] = _render_values_step(values, value, last)
        elif isinstance(value, str):
            template = jinja2.Template(value)
            new[key] = template.render(values)

            if last and new[key].find('{{') >= 0:
                raise Exception('j2 templates could not be fully resolved!')
        else:
            new[key] = value

    return new


import os
import yaml

def load_system_config(context):
    if os.path.exists(context['options']['imka_config']):
        with open(context['options']['imka_config']) as file:
            imkaConfig = yaml.safe_load(file)

        if not context['options']['imka_context']:
            context['options']['imka_context'] = imkaConfig['context']

        config = imkaConfig['contexts'][context['options']['imka_context']]

        context['value_files'] = config.get('value_files', [])
        context['hook_dirs'] = config.get('hook_dirs', [])

        context['options'] = config['options']


import hashlib
import docker

from . import util

from dataclasses import dataclass

@dataclass
class Config:
    configId: str
    version: str
    path: str
    docker_templating: bool
    content: str

    def apply(self, context):
        cli = context['docker_client']

        configs = cli.configs.list(filters={"name": self.version})

        if len(configs) == 0:
            return self.create(context)

        return False

    def create(self, context):
        cli = context['docker_client']

        if self.path:
            with util.open_with_context(context, self.path, 'rb') as file:
                data = file.read()

        if self.content:
            data = self.content.encode()

        labels = {
            'd4rk.io/imka/deployment': context['values']['deployment_fullname'],
            'd4rk.io/imka/config/id': self.configId,
        }

        templating = {}
        if self.docker_templating:
            templating = {
                'name': 'golang',
            }

        cli.configs.create(name=self.version, data=data, labels=labels, templating={})

        return self.version

    @staticmethod
    def down(context):
        cli = context['docker_client']

        configs = cli.configs.list(filters={"label": 'd4rk.io/imka/deployment={}'.format(context['values']['deployment_fullname'])})

        removed = []
        for config in configs:
            config.remove()
            removed.append(config.name)

        return removed

    def remove_old_versions(self, context):
        cli = context['docker_client']

        configs = cli.configs.list(filters={"label": 'd4rk.io/imka/config/id={}'.format(self.configId)})

        removed = []

        name = self.version
        for config in configs:
            if config.name != name:
                config.remove()
                removed.append(config.name)

        return removed

    @staticmethod
    def _get_id(context, path):
        localId = hashlib.sha256(path.encode()).hexdigest()[:32]

        return '{}-{}'.format(context['values']['deployment_fullname'], localId)

    @staticmethod
    def _get_version(context, path, config=None):
        if not config:
            with util.open_with_context(context, path) as file:
                config = file.read()

        return hashlib.sha256(config.encode()).hexdigest()

    @staticmethod
    def from_file(context, path, docker_templating):
        configId = Config._get_id(context, path)
        configVersion = Config._get_version(context, path)

        config = Config(configId, configVersion, path, docker_templating, None)
        context['configs'].append(config)

        return config

    @staticmethod
    def from_content(context, path, content):
        configId = Config._get_id(context, path)
        configVersion = Config._get_version(context, path, content)

        config = Config(configId, configVersion, None, False, content)
        context['configs'].append(config)

        return config

    @staticmethod
    def list(context):
        for config in context['configs'].values():
            print(config)

def apply_configs(context):
    for config in context['configs']:
        name = config.apply(context)
        if name:
            print('config {} created'.format(name))

def down_configs(context):
    removed = Config.down(context)
    for name in removed:
        print('config {} removed'.format(name))

def after_apply_configs(context):
    if context['options'].get('remove_old_config_versions_on_apply', False):
        for config in context['configs']:
            removed = config.remove_old_versions(context)
            for name in removed:
                print('config {} removed'.format(name))


from setuptools import setup

setup(
    name='imka',
    version='0.0.1',    
    description='A docker swarm package manager',
    url='https://github.com/byteplow/imka',
    author='byteplow',
    author_email='byteplow@posteo.de',
    license='MIT License',
    packages=['imka'],
    scripts=['command/imka'],
    install_requires=[
        'click',
        'dirhash',
        'docker',
        'Jinja2',
        'pyaml',
        'scp',
        'paramiko', 
    ],
    project_urls={
        'Source': 'https://github.com/byteplow/imka',
    },
)


import os
import yaml

from . import util

def load_frame(context):
    context['options']['frame_base_path'] = os.path.abspath(context['options']['frame'])

    with util.open_with_context(context, './frame.yml') as file:
        frameConfig = yaml.safe_load(file)

    context['values']['frame_name'] = frameConfig['name']
    context['compose_templates'] = frameConfig['compose_templates']

    valueFiles = os.path.join(context['options']['frame_base_path'], 'values.yml')
    if (os.path.exists(valueFiles)):
        context['value_files'] = [valueFiles] + context['value_files']

    hookDir = os.path.join(context['options']['frame_base_path'], 'hooks')
    if (os.path.exists(hookDir)):
        context['hook_dirs'] = [hookDir] + context['hook_dirs']



import os
import hashlib
import shutil

from . import util

from dirhash import dirhash
from dataclasses import dataclass
from paramiko import SSHClient, AutoAddPolicy
from scp import SCPClient

@dataclass
class Mount:
    mountId: str
    version: str
    path: str
    mountPath: str
    deployment: str

    def apply(self, context):
        provisioner = get_mount_provisioner(context)

        return provisioner.apply(context, self)

    @staticmethod
    def down(context):
        deployment = context['values']['deployment_fullname']
        provisioner = get_mount_provisioner(context)

        return provisioner.down(context, deployment)

    def remove_old_versions(self, context):
        provisioner = get_mount_provisioner(context)

        return provisioner.remove_old_versions(context, self)

    @staticmethod
    def _get_id(context, path):
        return hashlib.sha256(path.encode()).hexdigest()[:32]

    @staticmethod
    def _get_version(context, path):
        return dirhash(util.get_real_path(context, path), 'sha256')

    @staticmethod
    def from_path(context, path):
        mountId = Mount._get_id(context, path)
        mountVersion = Mount._get_version(context, path)
        deployment = context['values']['deployment_fullname']
        
        provisioner = get_mount_provisioner(context)
        mountPath = provisioner.get_mount_path(context, mountId, mountVersion, deployment)

        mount = Mount(mountId, mountVersion, path, mountPath, deployment)
        context['mounts'].append(mount)

        return mount

def apply_mounts(context):
    for mount in context['mounts']:
        names = mount.apply(context)
        for name in names:
            print('mount {} created'.format(name))

def after_apply_mounts(context):
    if context['options'].get('remove_old_mount_versions_on_apply', False):
        for mount in context['mounts']:
            removed = mount.remove_old_versions(context)
            for name in removed:
                print('mount {} removed'.format(name)) 

def down_mounts(context):
    removed = Mount.down(context)
    for name in removed:
        print('mount {} removed'.format(name))

def get_mount_provisioner(context):
    match context['options'].get('mount_provisioner', 'LocalMountProvisioner'):
        case 'SshMountProvisioner':
            return SshMountProvisioner()
        case 'LocalMountProvisioner':
            return LocalMountProvisioner()
        case _:
            return LocalMountProvisioner()

class LocalMountProvisioner:
    def get_mount_path(self, context, mountId, version, deployment):

        return os.path.join(self._get_versionless_path(context, mountId, deployment), version)
        
    def _get_idless_path(self, context, deployment): 
        basePath = context['options'].get('local_mount_provisioner_base_path', './tmp/mounts')

        return os.path.join(basePath, deployment)

    def _get_versionless_path(self, context, mountId, deployment):
        return os.path.join(self._get_idless_path(context, deployment), mountId)

    def apply(self, context, mount):
        if os.path.exists(mount.mountPath):
            return []

        return [self.create(context, mount)]

    def create(self, context, mount):
        shutil.copytree(util.get_real_path(context, mount.path), mount.mountPath)

        return mount.mountPath

    def remove_old_versions(self, context, mount):
        versionlessPath = self._get_versionless_path(context, mount.mountId, mount.deployment)

        removed = []

        for entry in os.listdir(versionlessPath):
            if entry != mount.version:
                path = os.path.join(versionlessPath, entry)
                shutil.rmtree(path)
                removed.append(path)

        return removed

    def down(self, context, deployment):
        idlessPath = self._get_idless_path(context, deployment)

        if os.path.exists(idlessPath):
            shutil.rmtree(idlessPath)
            return [idlessPath]

        return []

class SshMountProvisioner:
    def get_mount_path(self, context, mountId, version, deployment):
        return os.path.join(self._get_versionless_path(context, mountId, deployment), version)
        

    def _get_idless_path(self, context, deployment): 
        basePath = context['options']['ssh_mount_provisioner_base_path']

        return os.path.join(basePath, deployment)

    def _get_versionless_path(self, context, mountId, deployment):
        return os.path.join(self._get_idless_path(context, deployment), mountId)

    def _get_client(self, context, host):
        client = SSHClient()
        client.load_system_host_keys()
        client.set_missing_host_key_policy(AutoAddPolicy)
        client.connect(host['hostname'], port=host['port'], username=host['username'])

        return client

    def _path_exists(self, client, path):
        _, stdout, _ = client.exec_command('test -d {} && echo true'.format(path))

        return stdout.read() == b"true\n"

    def apply(self, context, mount):
        applied = []

        for host in context['options']['ssh_mount_provisioner_hosts']:
            client = self._get_client(context, host)

            if self._path_exists(client, mount.mountPath):
                continue

            applied.append(self.create(context, client, mount, host))

            client.close()

        return applied


    def create(self, context, client, mount, host):
        realPath = util.get_real_path(context, mount.path)

        client.exec_command('mkdir -p {}'.format(mount.mountPath))

        scp = SCPClient(client.get_transport())

        for entry in os.listdir(realPath):
            path = os.path.join(realPath, entry)
            scp.put(path, recursive=True, remote_path=mount.mountPath)

        scp.close()

        return '{}@{}:{}{}'.format(host['username'], host['hostname'], host['port'], mount.mountPath)

    def remove_old_versions(self, context, mount):
        versionlessPath = self._get_versionless_path(context, mount.mountId, mount.deployment)
        removed = []
        for host in context['options']['ssh_mount_provisioner_hosts']:
            client = self._get_client(context, host)

            _, stdout, _ = client.exec_command('for f in {}/*; do test "$f" != {} && rm -r $f && echo $f; done'.format(versionlessPath, mount.mountPath))

            for line in stdout.readlines():
                removed.append('{}@{}:{}{}'.format(host['username'], host['hostname'], host['port'], line.strip()))

            client.close()

        return removed

    def down(self, context, deployment):
        idlessPath = self._get_idless_path(context, deployment)
        removed = []
        for host in context['options']['ssh_mount_provisioner_hosts']:
            client = self._get_client(context, host)

            if self._path_exists(client, idlessPath):
                 _, stdout, _ = client.exec_command('rm -r {}'.format(idlessPath))
                 removed.append('{}@{}:{}{}'.format(host['username'], host['hostname'], host['port'], idlessPath))

            client.close()

        return removed